#########################################################################
""" remoteagent.py    -- Remote agent for Wing
                    
Copyright (c) 2015, Archaeopteryx Software, Inc.  All rights reserved.

Written by Stephan R.A. Deibel and John P. Ehresman

"""
#########################################################################

import sys
import os
import traceback

WINGHOME = os.path.dirname(os.path.abspath(__file__))
#sys.path.append(WINGHOME)
#import wingdbstub

def PrependBinDbgPath():
  """ Inserts WINGHOME/bin/dbg/src at the start of sys.path iff it
  exists.  It needs to be prepended as long as this file is in WINGHOME
  and the src/wingbase directory exists """
  
  bin_dbg = os.path.join(WINGHOME, 'bin', 'dbg', 'src')
  
  if os.path.exists(bin_dbg):
    sys.path.insert(0, bin_dbg)
    
PrependBinDbgPath()

def _TryCreateDir(dirname):
  if os.path.exists(dirname):
    return

  mergeimporter = _LoadFromWingbase(WINGHOME, 'mergeimporter')
  try:
    os.makedirs(dirname, mergeimporter.MergeDirImporter.CACHE_DIR_UMASK)
  except (IOError, OSError):
    print('Could not create directory: ' + repr(dirname) + '\n')
  else:
    print('Created directory: ' + repr(dirname) + '\n')
 
def _LoadFromWingbase(winghome, mod_name):
  """ Load a simple module from the wingbase directory.  Module must
  be in directory as a .py file and is loaded as a top-level module """

  import os.path
  if sys.hexversion >= 0x03070000:
    import importlib.util
    import importlib.machinery
  else:
    import imp
  
  dbg_subdir = 'bin/dbg'

  mod_base_name = mod_name + '.py'

  mod_full_path = None
  possible_subdir_list = [
    dbg_subdir + '/src/wingbase', 
    'src/wingbase', 
  ]
  for possible_subdir in possible_subdir_list:
    possible = os.path.join(winghome, possible_subdir, mod_base_name)
    if os.path.isfile(possible):
      mod_full_path = possible
      break
      
  if mod_full_path is None:
    raise ValueError("Could not find %s in winghome: %s" % (mod_base_name, winghome))
  
  if sys.hexversion >= 0x03070000:
    spec = importlib.machinery.PathFinder.find_spec(mod_name, [os.path.dirname(mod_full_path)])
    if spec is None:
      return None
    mod = importlib.util.module_from_spec(spec)
    if mod is None:
      return None
    spec.loader.exec_module(mod)
  else:
    fp, pathname, description = imp.find_module(mod_name, [os.path.dirname(mod_full_path)])
    try:
      mod = imp.load_module(mod_name, fp, pathname, description)
    finally:
      fp.close()
    
  return mod
 
def _GetPatchDirs(winghome, user_settings, wingversion):
  """Get list of (patch_name, fullpath) for all installed patches"""
  
  ver = '.'.join([str(s) for s in wingversion.kVersion])
  if wingversion.kBuild.startswith('b') or wingversion.kBuild.startswith('rc'):
    ver += wingversion.kBuild
  
  dirs = [
    os.path.join(winghome, 'patches', ver),
    os.path.join(user_settings, 'patches', ver)
  ]

  pair_list = []
  for patch_dir in dirs:
    try:
      name_list = os.listdir(patch_dir)
    except (IOError, OSError):
      name_list = []
    else:
      pair_list.extend([(name, os.path.join(patch_dir, name))
                        for name in name_list])
  pair_list.sort()

  return pair_list

def _CreateMetaImporter(winghome, user_settings, wingversion):
  """ Create meta importer to support .wingcode files and patches"""
  
  dbg_subdir = 'bin/dbg'

  mergeimporter = _LoadFromWingbase(winghome, 'mergeimporter')
  meta = mergeimporter.MergeDirImporter(logger=sys.stdout)
  print('Created meta importer')
  
  for name in ['debug', 'wingbase']:
    bin_dbg_dir = winghome + '/' + dbg_subdir + '/src/' + name
    src_name = winghome + '/src/' + name

    meta.add_dir(bin_dbg_dir, name)
    print('  Added dir', bin_dbg_dir)
    meta.add_dir(src_name, name)
    print('  Added dir', src_name)
    
  for name, dirname in _GetPatchDirs(winghome, user_settings, wingversion):
    possible_dir = os.path.join(dirname, 'bin', 'dbg', 'src')
    if os.path.isdir(possible_dir):
      meta.add_dir(possible_dir)
      print('  Added dir', possible_dir)

  return meta

def _SetupCacheDirs(winghome, meta, user_settings):
  """Set up cache dirs in meta importer, which we can only do
  after importing miscutils to find out the user settings dir"""
  
  if user_settings is None:
    return

  cache_dir = user_settings + '/debugger-pyc'
  _TryCreateDir(cache_dir)
    
  for name in ['debug', 'wingbase']:
    src_name = winghome + '/src/' + name
    sub_cache_dir = cache_dir + '/' + name
    _TryCreateDir(sub_cache_dir)
    meta.add_cache_dir(src_name, sub_cache_dir)

  return meta

# Hack to get version and assume Posix for the user settings directory since we 
# can't import miscutils yet; let WINGHOME/user-settings override default if present
# to support embedded devices with readonly main file system.
wingversion = _LoadFromWingbase(WINGHOME, 'wingversion')
user_settings = os.path.join(WINGHOME, 'user-settings')
if not os.path.exists(user_settings):
  if sys.platform == 'darwin':
    user_settings = os.path.expanduser('~/Library/Application Support/Wing IDE/v%i' % wingversion.kVersion[0])
  else:
    user_settings = os.path.join(os.path.expanduser('~'), '.wingide%i' % wingversion.kVersion[0])
meta = _CreateMetaImporter(WINGHOME, user_settings, wingversion)
sys.meta_path.insert(0, meta)

from wingbase import miscutils
# Setting kWingHome in miscutils here allows us to patch the module, where it
# then can't figure out where WINGHOME is on its own
miscutils.kWingHome = WINGHOME
_SetupCacheDirs(WINGHOME, meta, miscutils.kUserWingDir)
from wingbase import wingversion
wingversion.kPatches = _GetPatchDirs(WINGHOME, user_settings, wingversion)
from wingbase import remote

if __name__ == '__main__':

  # Tell spawn to wait longer for child processes to start since remote hosts
  # are often slow
  from wingbase import spawn
  spawn.PopenE.kForkTimeout = 20.0

  # Create wingdebugpw file
  miscutils.CreateInitialPasswordFile()
  
  # Set up log file
  user_dir = miscutils.kUserWingDir
  log_file = os.path.join(user_dir, 'remote-agent.log')
  log_obj = miscutils.CLoggedOutput(log_file, 1000000, truncate=False)
  sys.stderr = log_obj
  sys.stdout = log_obj
  
  print("Starting remote agent")
  print("PRODUCT=%s" % wingversion.kProduct)
  print("VERSION=%s-%s" % ('.'.join([str(v) for v in wingversion.kVersion]),
                           wingversion.kBuild))
  print("WINGHOME=%s" % miscutils._GetWingHome())
  print("WINGUSERDIR=%s" % user_dir)
  print("PYTHON=%s" % sys.executable)
  print(sys.version)
  print("ENVIRON=", os.environ)
  print("PATH=", sys.path)
  
  def failure():
    print("Usage: remoteagent.py --port=[hostname:]port [--basedir=basedir]")
    sys.exit(1)
    
  # Get host and port to connect to (host defaults to localhost)
  hostport = None
  basedir = None
  for arg in sys.argv[1:]:
    if arg.startswith('--port='):
      hostport = arg[len('--port='):]
    elif arg.startswith('--basedir='):
      basedir = arg[len('--basedir='):]
  print("hostport", hostport)
  print("basedir", basedir)
  
  if not hostport:
    failure()
    
  try:    
    if ':' in hostport:
      host, port = hostport.split(':')
      port = int(port)
      if host.strip() == '':
        host = '127.0.0.1'
    else:
      port = int(hostport)
      host = '127.0.0.1'
  except:
    failure()

  # Setup server and connect to IDE
  svr = None
  try:
    svr = remote.CRemoteServer(basedir)
    svr.Connect(host, port)
    svr.Run()
  except:
    if svr is None or not svr.fQuitting:
      print("UNEXPECTED EXCEPTION IN REMOTE AGENT")
    else:
      print("QUITTING REMOTE AGENT ON EXCEPTION")
    traceback.print_exc()
      
  